def escape(text: str) -> str:
    return text.replace('_', '\\_')


# underline
def uline(string: str, char: str = "="):
    return char * len(string)


# escape underline
def euline(string: str, char: str = "="):
    return uline(escape(string), char=char)


jinja_filters = dict(
    escape=escape,
    euline=euline,
    uline=uline,
)



from fnmatch import fnmatch
from functools import partial
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import Iterable, List, Optional, Union

from .model import Module

INIT_PY = '__init__.py'
MAIN_PY = '__main__.py'


def _path_to_str(path: Path, base_path: Path) -> str:
    path_str = str(path).replace(str(base_path.parent), '').replace('/', '.')
    if path_str.startswith('.'):
        path_str = path_str[1:]
    if path_str.endswith('.py'):
        path_str = path_str[:-3]
    return path_str


def _is_ignored(check_path: Path, ignore: List[str], base_path: Path) -> bool:
    path_str = _path_to_str(check_path, base_path=base_path)
    for ignored_item in ignore:
        if fnmatch(path_str, ignored_item):
            return True
    return False


def _get_submodules(base_path: Path, ignore: List[str]) -> Iterable[Module]:
    is_ignored = partial(_is_ignored, ignore=ignore, base_path=base_path)
    path_to_str = partial(_path_to_str, base_path=base_path)
    sub_modules = [
        p for p in sorted(base_path.rglob(INIT_PY)) if not is_ignored(p.parent)
    ]

    for sub_module in sub_modules:
        sub_module = sub_module.parent
        sub_sub_modules = [
            sub_sub_module
            for sub_sub_module in sub_module.glob('*.py')
            if sub_sub_module.name not in (INIT_PY, MAIN_PY)
            and not is_ignored(sub_sub_module)
        ]

        submodule_str = path_to_str(sub_module)

        yield Module(
            module_str=submodule_str,
            is_package=True,
            has_submodules=bool(len(sub_sub_modules)),
            parent_str='.'.join(submodule_str.split('.')[:-1]),
        )

        for sub_sub_module in sub_sub_modules:
            yield Module(
                module_str=path_to_str(sub_sub_module),
                is_package=False,
                has_submodules=False,
                parent_str=submodule_str,
            )


def get_module_hierarchy(
    name: Union[str, ModuleType], ignore: Optional[List[str]] = None
) -> Iterable[Module]:
    if ignore is None:
        ignore = []

    if isinstance(name, str):
        module = import_module(name)
    else:
        module = name

    assert isinstance(module, ModuleType) and module.__file__ is not None

    path = Path(module.__file__).parent

    yield from _get_submodules(path, ignore)



# pylint: disable=no-member
import io
import os
import shutil
from pathlib import Path
from typing import Dict, Optional

import docutils.io  # pylint: disable=import-error

from .mock import mock

base_path = os.getcwd()
virtual_files: Dict[str, str] = {}


def startstrip(search: str, pattern: str) -> Optional[str]:
    search, pattern = str(search), str(pattern)
    if search.startswith(pattern):
        return search[len(pattern) :]
    return None


def find_virtual(path: str) -> Optional[str]:
    match = startstrip(path, base_path)
    if match is not None and match[1:] in virtual_files:
        return virtual_files[match[1:]]
    return None


@mock(os.walk, module=os)
def os_walk(*args, **kwargs):
    for path, dirs, files in os.walk.original(*args, **kwargs):
        if path == base_path:
            files += [
                virtual_file_name.split('/')[-1] for virtual_file_name in virtual_files
            ]
        yield path, dirs, files


@mock(os.access, module=os)
def os_access(*args, **kwargs):
    if find_virtual(args[0]):
        return True

    return os.access.original(*args, **kwargs)


@mock(os.stat, module=os)
def os_stat(*args, **kwargs):
    if find_virtual(args[0]):
        return os.stat_result([0] * 10)

    return os.stat.original(*args, **kwargs)


@mock(open, module=docutils.io)
def docutils_io_opem(*args, **kwargs):
    file_match = find_virtual(args[0])

    if file_match:
        return io.TextIOWrapper(io.BytesIO(file_match.encode('utf-8')))

    return docutils.io.open.original(*args, **kwargs)


@mock(shutil.copyfile, module=shutil)
def shutil_copyfile(*args, **kwargs):
    file_match = find_virtual(args[0])
    if file_match:
        Path(args[1]).write_text(file_match, encoding="utf-8")
        return True
    return shutil.copyfile.original(*args, **kwargs)



from typing import Dict, List, Union

import jinja2

from .model import Module
from .util import jinja_filters

MODULE_TEMPLATE_STR = """
{{module|escape}} package
{{module|euline}}========
{% if subpackages %}
.. Subpackages
.. -----------

.. toctree::
   :maxdepth: 1
{% for subpackage in subpackages %}
   {{subpackage}}
{% endfor %}
{% endif %}
.. automodule:: {{module}}
   :members:
   :undoc-members:
   :show-inheritance:

{% if submodules %}
.. Submodules
.. ----------
{% for submodule in submodules %}
{{submodule|escape     }} module
{{submodule|euline('+')}}+++++++
.. automodule:: {{submodule}}
   :members:
   :undoc-members:
   :show-inheritance:
{% endfor %}
{% endif %}
"""


def prepare_jinja_template(template: str) -> jinja2.Template:
    env = jinja2.Environment()
    env.filters.update(jinja_filters)

    return env.from_string(template)


def prepare_rst_template_content(
    module: Module, module_hierarchy: List[Module]
) -> Dict[str, Union[str, List[str]]]:
    direct_submodules = [
        m for m in module_hierarchy if m.parent_str == module.module_str
    ]
    subpackages = [m.module_str for m in direct_submodules if m.is_package]
    submodules = [m.module_str for m in direct_submodules if not m.is_package]
    return dict(
        module=module.module_str, subpackages=subpackages, submodules=submodules
    )



# pylint: disable=import-error

from pathlib import Path

from sphinx.application import Sphinx
from sphinx.config import Config
from sphinx.util.logging import getLogger

from . import mock_filesystem
from .mock_filesystem import virtual_files
from .module_scanner import get_module_hierarchy
from .rst_preparation import (
    MODULE_TEMPLATE_STR,
    prepare_jinja_template,
    prepare_rst_template_content,
)

log = getLogger(__name__)


def process_text(data: str) -> str:
    return data


def config_intialized(app: Sphinx, config: Config):
    mock_filesystem.base_path = app.srcdir

    template = prepare_jinja_template(config.automagic_module_template)

    for module_to_process in config.automagic_modules:
        modules = list(
            get_module_hierarchy(module_to_process, ignore=config.automagic_ignore)
        )

        for module in modules:
            if not module.is_package:
                continue

            values = prepare_rst_template_content(module, modules)
            virtual_files[f"{module.module_str}.rst"] = template.render(**values)

    virtual_files.update(
        {
            file_name: process_text(data)
            for file_name, data in config.automagic_files.items()
        }
    )

    virtual_files.update(
        {
            file_name: Path(source_file_name).read_text(encoding='utf-8')
            for file_name, source_file_name in config.automagic_copy_files.items()
        }
    )

    log.info("virtual_files.keys() = %r", list(virtual_files.keys()))
    log.debug("virtual_files = %r", virtual_files)


def setup(app: Sphinx):
    app.connect('config-inited', config_intialized)
    app.add_config_value(
        name='automagic_module_template',
        default=MODULE_TEMPLATE_STR,
        rebuild='env',
        types=[str],
    )

    app.add_config_value(
        name='automagic_modules', default=[], rebuild='env', types=[list]
    )
    app.add_config_value(
        name='automagic_ignore', default=[], rebuild='env', types=[list]
    )
    app.add_config_value(
        name='automagic_files', default={}, rebuild='env', types=[dict]
    )
    app.add_config_value(
        name='automagic_copy_files', default={}, rebuild='env', types=[dict]
    )



import sys
from functools import wraps


def mock(original_function, module=None):
    def _mock(new_function):
        @wraps(new_function)
        def _inner(*args, **kwargs):
            return new_function(*args, **kwargs)

        _inner.original = original_function

        inner_module = module

        if inner_module is None:
            inner_module = sys.modules[original_function.__module__]
        setattr(inner_module, original_function.__name__, _inner)
        return _inner

    return _mock
